"""Awesome `pwnic` is a Python cli/package created with https://github.com/TezRomacH/python-package-template"""

import sys
import time
from importlib import metadata as importlib_metadata
from pathlib import Path

from pwnic.exploits import load_exploit


def get_version() -> str:
    try:
        return importlib_metadata.version(__name__)
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return "unknown"


version: str = get_version()

if sys.prefix == sys.base_prefix:
    print(
        "Your are not running inside a virtual environment! Be careful as this application might install other packages with pip"
    )
    time.sleep(5)

try:
    for type_dir in Path("exploits").iterdir():
        for exploit_dir in type_dir.iterdir():
            path = Path(exploit_dir)
            print(path)

            load_exploit(type=type_dir.stem, name=exploit_dir.stem)
except FileNotFoundError:
    pass



from typing import Any

import tomllib

from jsonschema import Draft202012Validator, validate


class _config_class:
    """
    Just a small wrapper so that config['prop'] and config.prop are equivalent
    """

    def __init__(self):
        with open("configs/config.toml", "rb") as fs:
            self._conf = tomllib.load(fs)
        self.schema = {
            "$schema": Draft202012Validator.META_SCHEMA["$id"],
            "type": "object",
            "properties": {
                "starting_timestamp": {"format": "time"},
                "ending_timestamp": {"format": "time"},
                "round_duration": {"type": "number", "exclusiveMinimum": 0},
                "team_id": {
                    "type": "integer",
                    "maximum": self._conf["highest_team_id"],
                    "minimum": 0,
                },
                "highest_team_id": {
                    "type": "integer",
                    "minimum": self._conf["team_id"],
                },
                "base_ip": {
                    "type": "string",
                },
                "exploit_warning_time": {
                    "type": "number",
                    "exclusiveMinimum": 0,
                },
                "submit_url": {"format": "uri"},
                "team_token": {"type": "string"},
                "log_level": {
                    "enum": ["DEBUG", "WARNING", "ERROR", "CRITICAL", "INFO"]
                },
                "flag_regex": {"type": "string", "format": "regex"},
                "timeout": {"type": "number", "exclusiveMinimum": 0},
                "ping_before_exploit": {"type": "boolean"},
            },
        }
        validate(
            self._conf,
            self.schema,
            format_checker=Draft202012Validator.FORMAT_CHECKER,
        )

    def __getattr__(self, __name: str) -> Any:
        return self._conf[__name]

    def __getitem__(self, __name: str) -> Any:
        return self._conf[__name]


config = _config_class()



from typing import Annotated

import os
import shutil
from pathlib import Path

from fastapi import FastAPI, File, Form, HTTPException, Response, UploadFile
from fastapi.responses import JSONResponse

import pwnic.db as db
from configs.Attacker import Attacker
from pwnic.exploits import exploits, load_exploit
from pwnic.submitter import Submitter
from pwnic.utils import install_package as package_install

app = FastAPI()


@app.post("/upload")
def upload_exploit(
    type: Annotated[str, Form()],
    name: Annotated[str, Form()],
    files: Annotated[list[UploadFile], File()],
) -> JSONResponse:
    path = f"exploits/{type}/{name}"

    os.makedirs(f"/tmp/{path}", exist_ok=True)
    for f in files:
        with open(f"/tmp/{path}/{f.filename}", "wb") as fs:
            fs.write(f.file.read())

    try:
        load_exploit(type, name, dir=f"/tmp/{path}")
    except ValueError:
        return JSONResponse(f"{type} is not valid", 422)
    except Exception:
        return JSONResponse("Unhandled Exception while loading exploit", 500)

    shutil.copytree(src=f"/tmp/{path}", dst=path, dirs_exist_ok=True)
    load_exploit(type, name)

    return JSONResponse({f.filename: f.size for f in files}, 201)


@app.get("/stats")
def get_stats():
    return db.get_stats()


@app.get("/install")
def install_package(package: str) -> dict[str, str]:
    if package_install(package):
        return {package: "installed"}
    else:
        return {package: "error"}


@app.get("/run")
def run_exploit(
    exploit: str, target: str, fetch_new_targets: bool = False
) -> Response:
    if exploit not in exploits:
        raise HTTPException(
            status_code=400,
            detail=f"{exploit} is not a valid exploit. It must be one of {exploits.keys()}",
        )
    if fetch_new_targets:
        Attacker.update_targets()

    Attacker.launch_attack(exploit, target)
    return JSONResponse("ok", 200)


@app.get("/launch_attacks")
def launch_exploits() -> None:
    Attacker.launch_all_attacks()


@app.get("/submit")
def submit_flags():
    flags = Submitter.fetch_unsent()
    out = Submitter.submit_flags(flags)
    print(out)

    if out is None:
        return JSONResponse("There was an error with the submit", 500)
    return out


@app.get("/addFlag")
def add_flag(flag: str) -> None:
    db.saveFlag(flag)


@app.get("/resetDB")
def reset_DB():
    os.remove("sqlite3.db")
    db.create_database()

    return {"ok": "database cleared"}



import re
import subprocess
from pathlib import Path


def install_package(package_name: str) -> bool:
    if re.search(r"[|&;\\]", package_name):
        return False
    package_name = re.sub(r"\s+", "", package_name)

    command = f"pip install {package_name}"
    print(f"running `{command}`")
    if subprocess.run(
        ["pip", "install", package_name], shell=False
    ).returncode:
        return False
    return True


def ensure_certs() -> tuple[str, str]:
    certdir = Path("cert")

    try:
        keyfile = next(certdir.glob("*key*"))
    except StopIteration:  # if no key, generate it
        print("Generating a key for self-signing")
        subprocess.run(
            args=[
                "openssl",
                "genpkey",
                "-algorithm",
                "RSA",
                "-out",
                "cert/key.pem",
                "-pkeyopt",
                "rsa_keygen_bits:4096",
            ],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        keyfile = next(certdir.glob("*key*"))

    try:  # check if there's a cert already
        certfile = next(certdir.glob("*cert*"))
    except StopIteration:  # if not, check for a key
        print("Generating a self-signed certificate")
        subprocess.run(
            args=[
                "openssl",
                "req",
                "-new",
                "-x509",
                "-sha256",
                "-key",
                "cert/key.pem",
                "-out",
                "cert/cert.pem",
                "-days",
                "30",
                "-subj",
                "/CN=pwnic",
            ],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        certfile = next(certdir.glob("*cert*"))

    return str(keyfile.resolve()), str(certfile.resolve())


def get_ip_addresses() -> list[str]:
    sub = subprocess.run(
        ["hostname", "--all-ip-addresses"], stdout=subprocess.PIPE
    )

    return sub.stdout.strip().decode().split("\n")



from typing import Any

from abc import abstractclassmethod
from datetime import datetime
from threading import Thread

from pwnic import db
from pwnic.config import config
from pwnic.exploits import exploits


class BaseAttacker:
    def __init__(cls) -> None:
        cls.started = False
        return

    @classmethod
    def launch_attack(cls, exploit: str, target_ip: str) -> None:
        target = db.get_targets(target_ip)
        target_id = 0
        t = Thread(
            target=exploits[exploit].run,
            args=(target_ip, target, target_id),
        )
        t.start()

    @classmethod
    def launch_all_attacks(cls) -> None:
        target = db.get_targets()
        target_ips: list[str] = [
            config["base_ip"].format(id=i)
            for i in range(1, config["highest_team_id"])
            if i != config["team_id"]
        ]

        for exploit in exploits.values():
            for ip in target_ips:
                exploit.run(ip, target)

        return

    @classmethod
    def update_targets(cls) -> None:
        new_targets = cls.fetch_targets()  # type: ignore
        db.add_targets(new_targets)

    @classmethod
    def has_game_started(cls) -> bool:
        if cls.started:
            return True

        start: datetime = config["starting_timestamp"]
        if datetime.now() > start:
            cls.started = True
            return True

        return False

    @abstractclassmethod
    def fetch_targets(cls) -> Any:
        pass



from typing import Any

import logging
from collections.abc import Iterable

import requests

from pwnic import db
from pwnic.config import config

log = logging.getLogger(__name__)


class Submitter:
    @classmethod
    def __init__(cls) -> None:
        return

    @classmethod
    def submit_flags(cls, flags: list[str]) -> Any | None:
        try:
            """
            this may be different for other CTFs
            """
            log.info(f"Submitting flags: {flags}")
            response: Any = requests.put(
                url=config.submit_url,
                headers={"X-Team-Token": config.team_token},
                json=flags,
                timeout=config["timeout"],  # nosec
            )  # .json()  # this is a list
            log.info(f"Got {response = }")
            return response.json()
        except requests.ConnectTimeout as e:
            log.exception("Gameserver didn't answer!")

        except requests.JSONDecodeError as e:
            log.critical(f"JSON decoding error!\n")
            raise e

        except Exception as e:
            log.exception(f"Unhandled exception was raised!")

        return None

    @classmethod
    def fetch_unsent(cls) -> list[str]:
        return db.fetch_unsent_flags()



import importlib
import sys
import time
from abc import ABC, abstractmethod
from pathlib import Path

from pwnic import db
from pwnic.utils import install_package


class Exploit(ABC):
    def __init__(self, name):
        self.name = name
        # self.log  # implement logger based on object instance (can i use the name?)
        return

    def register(self):
        exploits[self.name] = self

    @abstractmethod
    def attack(self, target_ip: str, target: dict | None = None) -> list[str]:
        pass

    def run(
        self,
        target_ip: str,
        target: dict | list | None = None,
        targets_id: int | None = None,
    ):
        t = time.perf_counter()

        flags: list[str] = self.attack(target_ip, target)

        elapsed_time = time.perf_counter() - t

        db.saveExploitRun(
            self.name, t, elapsed_time, target_ip, flags, targets_id
        )


class PythonExploit(Exploit):
    def __init__(
        self,
        name,
        path: str | Path | None = None,
        dir: str | Path | None = None,
    ):
        """
        path is the path to the file to import and to consider as entrypoint.
        dir is the path of the directory containing said file.
        If a path is specified and is valid, dir is ignored.
        If the specified path is a dir and no dir is specified than it is used as such.
        if both a path and a dir were specified but path is a dir, if they are equal then the dir is considered. If they are not, an exception is raised

        If the specified path is invalid or if no path is specified then `dir` is scanned looking for known entrypoints in order: `__main__.py`, `main.py`, `__init__.py`. In none can be found an error is raised.
        Default dir is exploits/{name}
        """

        super().__init__(name)
        if path is not None:
            if not Path(path).is_dir():
                self.path = Path(path)
                self.dir = Path(path).parent
            else:
                if dir is None:
                    self.dir = Path(path)
                elif path == dir:
                    self.dir = Path(dir)
                else:
                    raise Exception
                self.path = self.findPath(self.dir)

        else:
            if dir is not None:
                self.dir = Path(dir)
            else:
                self.dir = Path(f"exploits/python/{name}")
            self.path = self.findPath(self.dir)

        self.check_imports()

    def findPath(self, dir: str | Path) -> Path:
        if not Path(dir).is_dir():
            raise Exception(f"{Path(dir)} is not a directory")
        files = list(Path(dir).glob("**/[!_]*.py"))
        if len(files) == 1:
            return files[0]
        elif "__main__.py" in files:
            return Path(dir, "__main__.py")
        elif "main.py" in files:
            return Path(dir, "main.py")
        elif "__init__.py" in files:
            return Path(dir, "__init__.py")
        else:
            raise Exception(
                f"{self.name} is defined as a python exploit but has more than one .py file and none is '__main__.py', 'main.py' or '__init__.py'. \
            Please define an entrypoint."
            )

    def check_imports(self):
        retries = 1
        while retries >= 0:
            try:
                spec = importlib.util.spec_from_file_location(
                    self.name, self.path.resolve()
                )
                module = importlib.util.module_from_spec(spec)  # type: ignore
                sys.modules[self.name] = module
                spec.loader.exec_module(module)  # type: ignore
                self.exploit_function: function = getattr(
                    sys.modules[self.name], "main"
                )
                return
            except ModuleNotFoundError as e:
                modulename = (
                    e.msg.split()[-1].replace("'", "").replace('"', "")
                )
                install_package(modulename)
            except AttributeError:
                raise
            except Exception:
                retries -= 1

    def attack(self, target_ip: str, target: dict | None = None) -> list[str]:
        flags: list[str] = self.exploit_function(target_ip, target)
        return flags


def load_exploit(type: str, name: str, *args, **kwargs) -> Exploit:
    if type == "python":
        exploit = PythonExploit(name, *args, **kwargs)
    else:
        raise ValueError(
            f"{type} is not a valid exploit type.",
        )

    exploit.register()
    return exploit


exploits: dict[str, Exploit] = {}



import uvicorn

from pwnic.api import app
from pwnic.utils import ensure_certs, get_ip_addresses

keyfile, certfile = ensure_certs()
uvicorn.run(
    app,
    host=get_ip_addresses()[0],
    port=7777,
    ssl_keyfile=keyfile,
    ssl_certfile=certfile,
    log_level="trace",
)



import http
import logging
import logging.handlers

import colorlog
import requests
import urllib3
from rich.logging import RichHandler

from pwnic.config import config

log_level = getattr(logging, config.log_level)

# Create a logger
logger = logging.getLogger("pwnic")
logger.setLevel(logging.DEBUG)

# Create a file handler
file_handler = logging.handlers.RotatingFileHandler(  # 1 MiB max
    filename="logs/test.log", maxBytes=1048576, backupCount=3
)
file_handler.setLevel(log_level)

# Create a console handler
console_handler = RichHandler(
    rich_tracebacks=True, tracebacks_suppress=[requests, urllib3, http]
)
console_handler.setLevel(logging.DEBUG)


# Create a formatter and add it to the handlers
log_colors: dict[str, str] = colorlog.default_log_colors
log_colors["CRITICAL"] = "white,bg_red"
file_formatter = colorlog.ColoredFormatter(
    fmt="%(asctime)s | %(name)s | %(log_color)s%(levelname)s%(reset)s | %(message)s",
    log_colors=log_colors,
)

file_handler.setFormatter(file_formatter)

# Add the handlers to the logger
logger.addHandler(file_handler)
logger.addHandler(console_handler)



from typing import Any, Optional

import json
import re
import sqlite3
import time


def create_database():
    con = sqlite3.connect("sqlite3.db")
    cur = con.cursor()
    cur.execute(
        """CREATE TABLE IF NOT EXISTS targets(
        service NOT NULL,
        ip NOT NULL,
        target_json NOT NULL,
        used_by
        );"""
    )
    cur.execute(
        """CREATE TABLE IF NOT EXISTS runs(
        exploit NOT NULL,
        start_time NOT NULL,
        duration NOT NULL,
        target_ip NOT NULL,
        flags_json,
        n_flags NOT NULL,
        used_targets,
        FOREIGN KEY(used_targets) REFERENCES targets(rowid)
        )"""
    )
    cur.execute(
        """CREATE TABLE IF NOT EXISTS flags(
        flag PRIMARY KEY,
        status NOT NULL,
        run NOT NULL,
        FOREIGN KEY(run) REFERENCES runs(rowid)
        )
        """
    )


def add_targets(targets: dict[Any, Any]) -> int:
    con = sqlite3.connect("sqlite3.db")
    cur = con.cursor()
    for service in targets:
        for ip in targets[service]:
            cur.execute(
                "INSERT INTO targets (service, ip, target_json) values (?, ?, ?)",
                (service, ip, json.dumps(targets[service][ip])),
            )
    con.commit()
    auto_id = cur.lastrowid
    if auto_id is None:
        raise Exception("database error")
    return int(auto_id)


def get_targets(
    target_ip: str | None = None,
) -> list[Any]:  # TODO: The returned value needs fixing
    con = sqlite3.connect("sqlite3.db")
    cur = con.cursor()
    if target_ip:
        cur.execute(
            "SELECT targets.rowid, service, ip, target_json FROM targets JOIN runs ON targets.rowid == runs.used_targets WHERE ip == ?",
            (target_ip,),
        )
    else:
        cur.execute(
            "SELECT targets.rowid, service, ip, target_json FROM targets JOIN runs ON targets.rowid == runs.used_targets"
        )
    all = cur.fetchall()
    return all


def saveFlag(flag: str) -> None:
    con = sqlite3.connect("sqlite3.db")
    cur = con.cursor()
    cur.execute("INSERT INTO flags values (?, ?, ?)", (flag, "UNSENT", -1))
    con.commit()


def saveExploitRun(
    exploit_name: str,
    start_time: float,
    elapsed_time: float,
    target_ip: str,
    flags: list[str],
    used_targets_id: int,
) -> None:
    con = sqlite3.connect("sqlite3.db")
    cur = con.cursor()
    cur.execute(
        "INSERT INTO runs (exploit, start_time, duration, target_ip, flags_json, n_flags) values (?, ?, ?, ?, ?, ?)",
        (
            exploit_name,
            start_time,
            elapsed_time,
            target_ip,
            json.dumps(flags),
            len(flags),
        ),
    )
    # con.commit()
    runid = cur.lastrowid
    cur.execute(
        "UPDATE targets SET used_by = ? WHERE rowid = ? ",
        (runid, used_targets_id),
    )

    for flag in flags:
        try:
            cur.execute(
                "INSERT INTO flags values (?, ?, ?)",
                (flag, "UNSENT", runid),
            )
        except sqlite3.IntegrityError:
            pass  # we should probably log somewhere that we had a duplicate
    con.commit()


def fetch_unsent_flags() -> list[str]:
    con = sqlite3.connect("sqlite3.db")
    cur = con.cursor()
    cur.execute("SELECT flag FROM flags WHERE status == 'UNSENT'")
    return [x[0] for x in cur.fetchall()]


def get_stats():
    con = sqlite3.connect("sqlite3.db")
    cur = con.cursor()
    cur.execute("SELECT * FROM runs")
    return cur.fetchall()


create_database()
